"""
.. module:: LTL
   :synopsis: Represents the LTL language and provides model
              checking methods for it.

.. moduleauthor:: Alberto Casagrande <acasagrande@units.it>
"""

from .language import *
from .model_checking import modelcheck
from ..language import LNot

from .parser import Parser



def call_class_func(object, func_name):
    assert isinstance(func_name, str), 'func_name must be a string'
    if func_name not in object.__dict__:
        func = getattr(object, func_name, None)
        if func is not None:
            func()
            return
    print('object dont have function ' + func_name)



from pyModelChecking.CTLS import *
import unittest


def formula_into_str(phi):
    if isinstance(phi, bool):
        return str(phi).lower()
    return str(phi)


class TestFormulas(unittest.TestCase):
    def setUp(self):
        self.formulas = [True, 'p', AtomicProposition('q'), A(F(G('q'))),
                         G('p'), F('p'), X('p'), U('q', 'p'), R('q', 'p'),
                         A(Imply(F(G('q')), E(G('p')))),
                         A(G(Imply(And(Not('Close'), 'Start'),
                                   A(Or(G(Not('Heat')), F(Not('Error')))))))]

    def test_simplified_syntax(self):
        formulas = [(And(True, 'p'), Bool(True) & 'p'),
                    (Not(And(True, 'p')), ~(Bool(True) & 'p')),
                    (And(Not(True), 'p'), ~Bool(True) & 'p'),
                    (Or(True, Not('p')), Bool(True) | Not('p')),
                    (Or(True, Not('p')), True | Not('p')),
                    (Or(A(True), Not(X('p'))), A(True) | ~X('p'))]

        for old_syntax, new_syntax in formulas:
            self.assertEqual(old_syntax, new_syntax)

    def test_atomic_proposition(self):
        s = 'p'
        a = AtomicProposition(s)
        self.assertEqual(s, a)

        self.assertEqual(s, '{}'.format(a))

        with self.assertRaises(TypeError):
            AtomicProposition(1)

    def test_boolean(self):
        for b in [False, True]:
            a = Bool(b)
            self.assertEqual(b, a)
            self.assertEqual(a, b)

        with self.assertRaises(TypeError):
            Bool('a')

    def generic_test_unaryop(self, op, op_str, equivalent_restricted_op=None):
        i = 0
        for phi in self.formulas:
            s = '{}({})'.format(op_str, formula_into_str(phi))
            self.assertEqual(s, '{}'.format(op(phi)))
            if (equivalent_restricted_op is not None and
                    isinstance(phi, Formula)):
                self.assertEqual(op(phi).get_equivalent_restricted_formula(),
                                 equivalent_restricted_op(phi))

    def test_not(self):
        for phi in self.formulas:
            self.assertEqual('not {}'.format(formula_into_str(phi)),
                             '{}'.format(Not(phi)))
            if (isinstance(phi, Formula)):
                not_formula = Not(phi).get_equivalent_restricted_formula()
                lnot_formula = LNot(phi.get_equivalent_restricted_formula())
                self.assertEqual(not_formula, lnot_formula)

    def test_A(self):
        formula = (lambda phi:
                   Not(E(LNot(phi.get_equivalent_restricted_formula()))))
        self.generic_test_unaryop(A, 'A', formula)

    def test_E(self):
        formula = (lambda phi: E(phi.get_equivalent_restricted_formula()))
        self.generic_test_unaryop(E, 'E', formula)

    def test_G(self):
        formula = (lambda phi:
                   Not(F(LNot(phi)).get_equivalent_restricted_formula()))
        self.generic_test_unaryop(G, 'G', formula)

    def test_F(self):
        formula = (lambda phi:
                   U(True, phi.get_equivalent_restricted_formula()))
        self.generic_test_unaryop(F, 'F', formula)

    def test_X(self):
        formula = (lambda phi: X(phi.get_equivalent_restricted_formula()))
        self.generic_test_unaryop(X, 'X', formula)

    def generic_test_binaryop(self, op, middle_op_str,
                              equivalent_restricted_op=None):
        i = 0
        for phi in self.formulas:
            phi_str = formula_into_str(phi)
            for psi in self.formulas:
                s = '({} {} {})'.format(phi_str,
                                        middle_op_str,
                                        formula_into_str(psi))
                self.assertEqual(s, '{}'.format(op(phi, psi)))
                if (equivalent_restricted_op is not None and
                        isinstance(psi, Formula) and
                        isinstance(phi, Formula)):

                    formula = op(phi, psi).get_equivalent_restricted_formula()

                    self.assertEqual(formula,
                                     equivalent_restricted_op(phi, psi))

    def test_U(self):
        formula = (lambda phi, psi: U(phi.get_equivalent_restricted_formula(),
                                      psi.get_equivalent_restricted_formula()))
        self.generic_test_binaryop(U, 'U', formula)

    def test_R(self):
        formula = (lambda phi, psi:
                   Not(U(LNot(phi.get_equivalent_restricted_formula()),
                         LNot(psi.get_equivalent_restricted_formula()))))
        self.generic_test_binaryop(R, 'R', formula)

    def test_or(self):
        formula = (lambda phi, psi:
                   Or(phi.get_equivalent_restricted_formula(),
                      psi.get_equivalent_restricted_formula()))
        self.generic_test_binaryop(Or, 'or', formula)

    def test_and(self):
        formula = (lambda phi, psi:
                   Not(Or(LNot(phi.get_equivalent_restricted_formula()),
                          LNot(psi.get_equivalent_restricted_formula()))))
        self.generic_test_binaryop(And, 'and', formula)

    def test_imply(self):
        formula = (lambda phi, psi:
                   Or(LNot(phi.get_equivalent_restricted_formula()),
                      psi.get_equivalent_restricted_formula()))
        self.generic_test_binaryop(Imply, '-->', formula)


if __name__ == '__main__':
    unittest.main()



"""
.. module:: CTL.model_checking
   :synopsis: Provides model checking methods for the CTL language.

.. moduleauthor:: Alberto Casagrande <acasagrande@units.it>
"""

from .language import *
from trainify.validator.pyModelChecking.graph import compute_SCCs
from trainify.validator.pyModelChecking.kripke import Kripke

import trainify.validator.pyModelChecking.CTLS

from .parser import Parser

import sys

CTLS = sys.modules['trainify.validator.pyModelChecking.CTLS']


def _init_formula_in(formula, L):
    if formula not in L:
        L[formula] = set()


def _checkAtomicProposition(kripke, formula, L):
    if formula not in L:
        Lformula = set()
        for v in kripke.states():
            if formula.name in kripke.labels(v):
                Lformula.add(v)

        L[formula] = Lformula

    return L[formula]


def _checkNot(kripke, formula, L):
    if formula not in L:
        Lphi = _checkStateFormula(kripke, formula.subformula(0), L)

        Lformula = set()
        for v in kripke.states():
            if v not in Lphi:
                Lformula.add(v)

        L[formula] = Lformula

    return L[formula]


def _checkEX(kripke, formula, L):
    if formula not in L:

        p_formula = formula.subformula(0)
        Lphi = _checkStateFormula(kripke, p_formula.subformula(0), L)

        Lformula = set()
        for (src, dst) in kripke.transitions_iter():
            if dst in Lphi:
                Lformula.add(src)

        L[formula] = Lformula

    return L[formula]


def _checkOr(kripke, formula, L):
    if formula not in L:
        Lformula = set()
        for sf in formula.subformulas():
            for v in _checkStateFormula(kripke, sf, L):
                Lformula.add(v)

        L[formula] = Lformula

    return L[formula]


def _checkEU(kripke, formula, L):
    if formula not in L:
        Lphi = []
        p_formula = formula.subformula(0)
        for i in range(2):
            Lphi.append(_checkStateFormula(kripke, p_formula.subformula(i), L))

        subgraph = kripke.get_subgraph(Lphi[0])
        subgraph = subgraph.get_reversed_graph()

        for v in Lphi[0]:
            for w in (kripke.next(v) & Lphi[1]):
                try:
                    subgraph.add_edge(w, v)
                except Exception:
                    pass

        # Some of the nodes satisfying p_formula.subformula(1)
        # can be non-reachable from the reversed subgraph of
        # the nodes satisfying p_formula.subformula(0).
        # Hence, the former must be added to the
        # investigated subgraph
        for v in Lphi[1] - subgraph.nodes():
            subgraph.add_node(v)

        L[formula] = subgraph.get_reachable_set_from(Lphi[1])

    return L[formula]


def _checkEG(kripke, formula, L):
    if formula not in L:
        p_formula = formula.subformula(0)
        Lphi = _checkStateFormula(kripke, p_formula.subformula(0), L)

        subgraph = kripke.get_subgraph(Lphi)
        subgraph = subgraph.get_reversed_graph()
        SCCs = compute_SCCs(subgraph)

        T = set()
        for scc in SCCs:
            v = next(iter(scc))
            if len(scc) > 1 or v in subgraph.next(v):
                T.update(scc)

        L[formula] = subgraph.get_reachable_set_from(T)

    return L[formula]


def _checkStateFormula(kripke, formula, L):
    if isinstance(formula, CTLS.Not):
        return _checkNot(kripke, formula, L)

    if isinstance(formula, CTLS.Or):
        return _checkOr(kripke, formula, L)

    if (isinstance(formula, CTLS.Bool) or isinstance(formula, bool)):
        Lang = sys.modules[formula.__module__]
        if formula == Bool(True):
            Lformula = set(kripke.states())
            L[Lang.Bool(True)] = Lformula
        else:
            Lformula = set()
            L[Lang.Bool(False)] = Lformula

        return Lformula

    if isinstance(formula, CTLS.AtomicProposition):
        return _checkAtomicProposition(kripke, formula, L)

    if isinstance(formula, CTLS.E):
        p_formula = formula.subformula(0)
        if isinstance(p_formula, CTLS.G):
            return _checkEG(kripke, formula, L)

        if isinstance(p_formula, CTLS.U):
            return _checkEU(kripke, formula, L)

        if isinstance(p_formula, CTLS.X):
            return _checkEX(kripke, formula, L)

    restr_f = formula.get_equivalent_restricted_formula()

    Lalter_formula = _checkStateFormula(kripke, restr_f, L)

    L[formula] = Lalter_formula

    return Lalter_formula


def modelcheck(kripke, formula, parser=None, F=None):
    ''' Model checks any CTL formula on a Kripke structure.

    This method performs CTL model checking of a formula on a given
    Kripke structure.

    :param kripke: a Kripke structure.
    :type kripke: Kripke
    :param formula: the formula to model check.
    :type formula: a type castable in a CTL.Formula or a string representing
                   a CTL formula
    :param parser: a parser to parse a string into a CTL.Formula.
    :type parser: CTL.Parser
    :param F: a list of fair states
    :type F: Container
    :returns: a list of the Kripke structure states that satisfy the formula.
    '''

    if isinstance(formula, str):
        if parser is None:
            parser = Parser()
        formula = parser(formula)

    if not isinstance(formula, Formula):
        try:
            formula = formula.cast_to(sys.modules[__name__])
        except Exception:
            raise TypeError('expected a CTL state formula, ' +
                            'got {}'.format(formula))

    if not isinstance(formula, StateFormula):
        raise TypeError('expected a CTL state formula, got {}'.format(formula))

    if not isinstance(kripke, Kripke):
        raise TypeError('expected a Kripke structure, got {}'.format(kripke))

    if F is not None:
        kripke = kripke.clone()

        fair_label = kripke.label_fair_states(F)

        formula = formula.get_equivalent_non_fair_formula(fair_label)

    return _checkStateFormula(kripke, formula, L=dict())



from pyModelChecking.BDD import *

import unittest


class TestOBDD(unittest.TestCase):
    def setUp(self):
        self.a = BDDNode('a', BDDNode(0), BDDNode(True))
        self.b = BDDNode('a', BDDNode(True), BDDNode(False))
        self.c = BDDNode('c', self.b, BDDNode(False))
        self.d = BDDNode('b', BDDNode(1), BDDNode(0))

        self.ordering = ['c', 'a']

    def test_BDD(self):
        for (bdd, text, vars) in [(self.a, 'a', ['a']),
                                  (self.b, '~a', ['a']),
                                  (self.c, '~c & ~a', ['a', 'c']),
                                  (self.d, '~b', ['b'])]:

            self.assertEqual('{}'.format(bdd), text)
            self.assertEqual(bdd.variables(), set(vars))

        get_ancestors = (lambda: ['{}'.format(BDD) for BDD in BDDNode.nodes()])

        before_e = get_ancestors()

        e = BDDNode('r', BDDNode('c', self.d, self.a),
                    BDDNode('h', self.a, self.d))

        with_e = get_ancestors()

        e = None

        after_e = get_ancestors()

        self.assertEqual(before_e, after_e)
        self.assertNotEqual(with_e, after_e)

    def test_OBDD(self):
        oa = OBDD(self.a, self.ordering)
        ob = OBDD(self.c, self.ordering)

        tests_list = [(oa, 'lambda c,a: a', ['a']),
                      (ob, 'lambda c,a: ~c & ~a', ['a', 'c']),
                      (ob.restrict('c', False), 'lambda c,a: ~a', ['a']),
                      (ob.restrict('c', 1), 'lambda c,a: 0', []),
                      (ob.restrict('a', 0), 'lambda c,a: ~c', ['c']),
                      (ob.restrict('a', True), 'lambda c,a: 0', []),
                      (~ob, 'lambda c,a: (~c & a) | (c)', ['a', 'c']),
                      (oa & ob, 'lambda c,a: 0', []),
                      (oa | ob, 'lambda c,a: (~c) | (c & a)', ['a', 'c']),
                      (~(oa & ob) | (oa | ob), 'lambda c,a: 1', [])]

        for (obdd, text, vars) in tests_list:
            self.assertEqual('{}'.format(obdd), text)
            self.assertEqual(obdd.variables(), set(vars))

        self.assertEqual(ob.restrict('a', True), 0)
        self.assertEqual(~(oa & ob) | (oa | ob), 1)

        with self.assertRaises(RuntimeError):
            OBDD(self.d, self.ordering)

    def test_OBDD_parser(self):
        oa = OBDD(self.a, self.ordering)
        ob = OBDD(self.c, self.ordering)

        for obdd in [oa, ob, ~ob, oa & ob,  oa | ob,  ~(oa & ob) | (oa | ob)]:
            self.assertEqual(OBDD('{}'.format(obdd.root), self.ordering),
                             obdd)

        for (a, b) in [('a & ~a', '0'),
                       ('a | ~a', '1'),
                       ('~a | c', '~(a & ~c)'),
                       ('~~~~a', 'a'),
                       ('~(~a|~c)', 'a&c')]:
            self.assertEqual(OBDD(a, self.ordering), OBDD(b, self.ordering))

        with self.assertRaises(RuntimeError):
            OBDD('a|~b', self.ordering)

        with self.assertRaises(RuntimeError):
            OBDD('lambda c,a: a|~b')


if __name__ == '__main__':
    unittest.main()



from weakref import WeakSet
from .ordering import Ordering


def BDDsons_and_BDD(operator, A, B, ordering, r_cache):
    low = apply(operator, A.low, B, ordering, r_cache)
    high = apply(operator, A.high, B, ordering, r_cache)
    return BDDNonTerminalNode(A.var, low, high)


def BDD_and_BDDsons(operator, A, B, ordering, r_cache):
    low = apply(operator, A, B.low, ordering, r_cache)
    high = apply(operator, A, B.high, ordering, r_cache)
    return BDDNonTerminalNode(B.var, low, high)


def BDDsons_and_BDDsons(operator, A, B, ordering, r_cache):
    low = apply(operator, A.low, B.low, ordering, r_cache)
    high = apply(operator, A.high, B.high, ordering, r_cache)
    return BDDNonTerminalNode(A.var, low, high)


def apply(operator, A, B, ordering, r_cache):
    if A not in r_cache:
        r_cache[A] = dict()

    if B in r_cache[A]:
        return r_cache[A][B]

    r_cache[A][B] = compute(operator, A, B, ordering, r_cache)

    return r_cache[A][B]


def compute(operator, A, B, ordering, r_cache):
    if isinstance(A, BDDTerminalNode):
        if isinstance(B, BDDTerminalNode):
            return BDDTerminalNode(operator(A.value, B.value))

        return BDD_and_BDDsons(operator, A, B, ordering, r_cache)

    if (isinstance(B, BDDTerminalNode) or ordering.in_order(A.var, B.var)):
        return BDDsons_and_BDD(operator, A, B, ordering, r_cache)

    if A.var == B.var:
        return BDDsons_and_BDDsons(operator, A, B, ordering, r_cache)

    if ordering.in_order(B.var, A.var):
        return BDD_and_BDDsons(operator, A, B, ordering, r_cache)

    raise RuntimeError('Unsupported configuration %s %s' % A,  B)


def descendents(root, checked=None):
    if checked is None:
        checked = set()

    desc = set()
    stack = [root]

    while stack:
        node = stack.pop()
        if node not in desc and node not in checked:
            desc.add(node)

            if (isinstance(node, BDDNonTerminalNode)):
                stack.append(node.low)
                stack.append(node.high)

    return desc


def ancestors(leaf, checked=None):
    if checked is None:
        checked = set()

    anc = set()
    stack = [leaf]

    while stack:
        node = stack.pop()
        if node not in anc and node not in checked:
            anc.add(node)

            stack.extend((node.f_low | node.f_high-checked)-anc)

    return anc


class BDDNode(object):
    '''
    A class to represent Binary Decision Diagram (BDD) nodes.
    '''
    def __new__(cls, *data):
        if len(data) == 1:
            return BDDTerminalNode(data[0])

        if len(data) == 3:
            return BDDNonTerminalNode(*data)

        raise RuntimeError('passed {} parameters, '.format(len(data)) +
                           'but only  BDDNode(value) for terminal node and ' +
                           'BDDNode(var, low, high) for non terminal node ' +
                           'have been implemented')

    def __hash__(self):
        ''' Compute a hash for a BDDNode

        :returns: the id of `self`
        :rtype: int
        '''
        return id(self)

    def restrict(self, var, value):
        ''' Partially evaluate the binary function encoded by a BDD.

        :param var: name of the variable to be set
        :type var: str
        :param value: value to be set
        :type value: bool or int
        :returns: the BDDNode representing the partial evaluation of
                  :math:`f` with :param var: set to :param value: where
                  :math:`f` is the function encoded by the object
        :rtype: BDDNode
        '''
        if isinstance(value, int):
            if value == 1:
                value = True
            if value == 0:
                value = False

        if not (isinstance(var, str) and isinstance(value, bool)):
            raise TypeError('expected a pair (str, bool), ' +
                            'got ({}, {})'.format(var.__class__,
                                                  value.__class__))

        return cache_restrict(self, var, value, dict())

    def __reset__(self):
        self.f_low = WeakSet()
        self.f_high = WeakSet()

    def descendents(self):
        ''' Computes the descendents of this node.

        :returns: the set of nodes that have the current object as ancestor
        :rtype: set of :class:`BDDNode`
        '''
        return descendents(self, set())

    def ancestors(self):
        ''' Computs the ancestors of this node.

        :returns: the set of nodes that have the current object as descendent
        :rtype: set of :class:`BDDNode`
        '''
        return ancestors(self, set())

    @staticmethod
    def nodes():
        ''' Return all the BDD nodes stored in memory.

        :returns: all the BDD nodes that are stored in memory
        :rtype: set of `BDDNode`
        '''
        return BDDNode(0).ancestors() | BDDNode(1).ancestors()

    def variables(self):
        ''' Return the variables contained into a BDD.

        :returns: all the variable that label some the descendents of this node
        :rtype: set of str
        '''
        return set([node.var for node in self.descendents()
                    if isinstance(node, BDDNonTerminalNode)])

    def __repr__(self):
        return self.__str__()


def find_isomorph(var, low, high):
    if len(low.f_low) < len(high.f_high):
        node_set = low.f_low
        lh_test = (lambda low, high: high is node.high)
    else:
        node_set = high.f_high
        lh_test = (lambda low, high: low is node.low)

    for node in node_set:
        if (isinstance(node, BDDNonTerminalNode) and var == node.var and
                lh_test(low, high)):
            return node

    return None


def cache_restrict(bdd, var, value, r_cache):
    if bdd in r_cache:
        return r_cache[bdd]

    r_cache[bdd] = compute_restrict(bdd, var, value, r_cache)

    return r_cache[bdd]


def compute_restrict(bdd, var, value, r_cache):
    if isinstance(bdd, BDDTerminalNode):
        return bdd

    if bdd.var == var:
        if value:
            return cache_restrict(bdd.high, var, value, r_cache)
        else:
            return cache_restrict(bdd.low, var, value, r_cache)
    else:
        low = cache_restrict(bdd.low, var, value, r_cache)
        high = cache_restrict(bdd.high, var, value, r_cache)

        return BDDNonTerminalNode(bdd.var, low, high)


class BDDNonTerminalNode(BDDNode):
    def __new__(cls, var, low, high):
        for p in [low, high]:
            if not isinstance(p, BDDNode):
                raise TypeError('expected an BBD node, got {}'.format(p))

        if low is high:
            return low

        node = find_isomorph(var, low, high)
        if node is not None:
            return node

        node = super(BDDNode, cls).__new__(cls)
        node.__reset__(var, low, high)

        return node

    def respect_ordering(self, O, checked=None):
        ''' Test whether a BDDNode respects an ordering.

        :param O: the ordering to be tested
        :type O: Ordering
        :param checked: already checked nodes
        :type checked: set
        :returns: True if and only if the variable of `node` is smaller that
                  those of `node.low` and `node.high` for any node in the
                  subgraph rooted in this object
        :rtype: bool
        '''
        if not isinstance(O, Ordering):
            TypeError('expected an Ordering, got {}'.format(O))

        if checked is None:
            checked = set()

        if self in checked:
            return True

        if self.var not in O:
            raise RuntimeError('%s in not in %s' % (self.var, O))

        for son in [self.low, self.high]:
            if (isinstance(son, BDDNonTerminalNode) and
                    not O.in_order(self.var, son.var)):
                return False

        if (self.high.respect_ordering(O, checked) and
                self.low.respect_ordering(O, checked)):
            checked.add(self)

            return True

        return False

    def __reset__(self, var, low, high):
        ''' Reset the value of a BDDNonTerminalNode.

        :param var: variable name
        :type var: str
        :param low: the low BDDNode
        :type low: BDDNode
        :param high: the high BDDNode
        :type high: BDDNode
        '''
        super(BDDNonTerminalNode, self).__reset__()

        self.var = var
        self.low = low
        self.high = high

        self.low.f_low.add(self)
        self.high.f_high.add(self)

    def __invert__(self, r_cache=None):
        ''' Invert the binary function represented by a BDDNode.

        :param r_cache: a dictionary that caches already computed results
        :type r_cache: dict or None
        :returns: the BDDNode representing the function `1-f` where `f` is the
                  function depicted by `self`
        :rtype: BDDNode
        '''
        if r_cache is None:
            r_cache = dict()

        if self in r_cache:
            return r_cache[self]

        r_cache[self] = BDDNonTerminalNode(self.var,
                                           self.low.__invert__(r_cache),
                                           self.high.__invert__(r_cache))
        return r_cache[self]

    def __hash__(self):
        return id(self)

    def __eq__(self, O):
        ''' Test the equivalence of two BDDs.

        :param O: a BDD node
        :type O: BDDNode
        :returns: `True` if the two BDDs rooted in `self` and `O` are
                  isomorph,  `False`,  otherwise
        :rtype: bool
        '''
        return self is O

    def __str__(self):
        ''' Produce a string that represents a BDD.

        :returns: a string that represents the BDD rooted in `self`
        :rtype: str
        '''
        repr = []

        for succ,  neg in [(self.low, '~'), (self.high, '')]:
            if isinstance(succ, BDDTerminalNode):
                if succ.value:
                    repr.append('%s%s' % (neg, self.var))
            else:
                repr.append('%s%s & %s' % (neg, self.var, succ))

        if len(repr) == 2:
            return '(%s) | (%s)' % (repr[0], repr[1])

        return repr[0]


class BDDTerminalNode(BDDNode):
    Tnodes = {}

    def __new__(cls, value):
        if not (value in set([0, 1, False, True])):
            raise TypeError('expected a value among [0, 1, False, True], ' +
                            'got {}'.format(value))

        if value not in BDDTerminalNode.Tnodes:
            node = super(BDDNode, cls).__new__(cls)

            # reset the node and set the node value
            node.__reset__(value)

            BDDTerminalNode.Tnodes[value] = node

        return BDDTerminalNode.Tnodes[value]

    def respect_ordering(self, O, checked=None):
        ''' Test whether a BDDNode respects an Ordering or not.

        :param O: the ordering to be tested
        :type O: Ordering
        :param checked: already checked nodes
        :type checked: set
        :returns: True
        :rtype: bool
        '''
        return True

    def __reset__(self, value):
        ''' Reset the value of a BDDTerminalNode.

        :param value: the new value
        :type value: bool
        '''
        super(BDDTerminalNode, self).__reset__()
        self.value = value

    def __invert__(self, r_cache=None):
        ''' Invert the binary function represented by a BDDNode.

        :param r_cache: a dictionary that caches already computed results
        :type r_cache: dict or None
        :returns: the BDDNode representing the function `1-f` where `f` is the
                  function depicted by `self`
        :rtype: BDDNode
        '''
        if r_cache is None:
            r_cache = dict()

        if self in r_cache:
            return r_cache[self]

        r_cache[self] = BDDTerminalNode(not self.value)

        return r_cache[self]

    def __hash__(self):
        return id(self)

    def __eq__(self, O):
        ''' Test the equivalence of two BDDs.

        :param O: a BDD node
        :type O: BDDNode
        :returns: `True` if the two BDDs rooted in `self` and `O` are
                  isomorph,  `False`,  otherwise
        :rtype: bool
        '''
        return self is O

    def __str__(self):
        if self.value:
            return '1'
        return '0'



"""
.. module:: LTL.model_checking
   :synopsis: Provides model checking methods for the LTL language.

.. moduleauthor:: Alberto Casagrande <acasagrande@units.it>
"""

from .language import *
from trainify.validator.pyModelChecking.graph import DiGraph
from trainify.validator.pyModelChecking.graph import compute_SCCs
from trainify.validator.pyModelChecking.kripke import Kripke
from trainify.validator.pyModelChecking.CTLS import LNot as LNot

from .parser import Parser

import sys

# if 'pyModelChecking.CTLS' not in sys.modules:
import trainify.validator.pyModelChecking.CTLS

LTL = sys.modules[__name__]
CTLS = sys.modules['trainify.validator.pyModelChecking.CTLS']


def _get_closure(formula):
    closure = set()
    T = [formula]

    Lang = sys.modules[formula.__module__]
    while len(T) > 0:
        phi = T.pop()
        if phi not in closure:
            closure.add(phi)
            T.append(LNot(phi))
            if isinstance(phi, CTLS.X):
                T.append(phi.subformula(0))
            else:
                if (isinstance(phi, CTLS.Not) and
                        isinstance(phi.subformula(0), CTLS.X)):
                    sf = phi.subformula(0).subformula(0)
                    T.append(Lang.X(LNot(sf)))
                else:
                    if isinstance(phi, CTLS.Or):
                        for sf in phi.subformulas():
                            T.append(sf)
                    else:
                        if isinstance(phi, CTLS.U):
                            T.append(phi.subformula(0))
                            T.append(phi.subformula(1))
                            T.append(Lang.X(phi))
                        else:
                            if not (isinstance(phi, CTLS.Not) or
                                    isinstance(phi, CTLS.AtomicProposition) or
                                    isinstance(phi, CTLS.Bool)):
                                raise TypeError('expected a LTL path ' +
                                                'formula  restricted to ' +
                                                '"or",  "not",  ' +
                                                '"U" and "X", got ' +
                                                '{}'.format(phi))

    return closure


class _TableuAtom(set):
    def __init__(self, state, formulas=None):
        self.state = state
        if (formulas is None):
            formulas = set()

        super(_TableuAtom, self).__init__(formulas)

    def __or__(self, A):
        if (isinstance(A, _TableuAtom)):
            if self.state != A.state:
                raise RuntimeError('{} and {}'.format(self, A) +
                                   ' have different states')
            return self | set(A)

        return _TableuAtom(self.state, super(_TableuAtom, self).__or__(A))

    def __ror__(self, A):
        return self | A

    def __and__(self, A):
        if (isinstance(A, _TableuAtom)):
            if self.state != A.state:
                raise RuntimeError('{} and {}'.format(self, A) +
                                   ' have different states')
            return self & set(A)

        return _TableuAtom(self.state, super(_TableuAtom, self).__and__(A))

    def __rand__(self, A):
        return self & A

    def clone(self):
        return _TableuAtom(self.state, set(self))

    def __str__(self):
        return '(%s, %s)' % (self.state, set(self))

    def __repr__(self):
        return str(self)


def _does_respect_Xs(Xs_in_closure, s_atom, d_atom):
    for phi in Xs_in_closure:
        if (phi.subformula(0) in d_atom) ^ (phi in s_atom):
            return False

    return True


def _build_state_dict(atoms):
    state_dict = {}
    for i in range(len(atoms)):
        if atoms[i].state not in state_dict:
            state_dict[atoms[i].state] = [i]
        else:
            state_dict[atoms[i].state].append(i)

    return state_dict


def _build_atoms(K, closure):
    A = []
    for state in K.states():
        A.append(_TableuAtom(state))

    # this is to avoid issues with the "not X" case
    cl_list = sorted(list(closure), key=(lambda a: a.height
    if not (isinstance(a, CTLS.Not) and
            isinstance(a.subformula(0), CTLS.X))
    else a.height - 1))

    for phi in cl_list:
        Lang = sys.modules[phi.__module__]

        if phi != Lang.Not(True) and phi != Lang.Bool(False):
            neg_phi = LNot(phi)

            A_tail = []
            if isinstance(phi, CTLS.Bool):
                for atom in A:
                    atom.add(phi)
            else:
                if isinstance(phi, CTLS.AtomicProposition):
                    for atom in A:
                        if phi in K.labels(atom.state):
                            atom.add(phi)
                        else:
                            atom.add(neg_phi)

            if (isinstance(phi, CTLS.Or)):
                sf = phi.subformulas()

                for atom in A:
                    if sum([f in atom for f in sf]):
                        atom.add(phi)
                    else:
                        atom.add(neg_phi)

            if (isinstance(phi, CTLS.Not) and
                    isinstance(phi.subformula(0), CTLS.X)):
                sf = phi.subformula(0).subformula(0)

                for atom in A:
                    if phi.subformula(0) not in atom:
                        if phi not in atom:
                            new_atom = atom | set([phi, Lang.X(LNot(sf))])
                            A_tail.append(new_atom)
                            atom.add(phi.subformula(0))
                        else:
                            atom.add(Lang.X(LNot(sf)))

            if isinstance(phi, CTLS.U):
                sf = phi.subformulas()

                for atom in A:
                    if (sf[1] in atom):
                        atom.add(phi)
                        A_tail.append(atom | set([Lang.X(phi)]))
                    else:
                        if (sf[0] in atom):
                            A_tail.append(atom | set([phi, Lang.X(phi)]))
                        atom.add(neg_phi)
                    atom.add(Lang.Not(Lang.X(phi)))

            A.extend(A_tail)

            for atom in A:
                if phi not in atom and neg_phi not in atom:
                    A.append(atom | set([phi]))
                    atom.add(neg_phi)

    return A


class _Tableu(DiGraph):
    def __init__(self, K, formula=None, closure=None):
        if closure is None:
            if formula is None:
                raise RuntimeError('either closure or formula ' +
                                   'must be provided')
            closure = _get_closure(formula)

        Xs_in_closure = set([p for p in closure if isinstance(p, CTLS.X)])

        self.atoms = _build_atoms(K, closure)
        state_dict = _build_state_dict(self.atoms)

        super(_Tableu, self).__init__(V=range(len(self.atoms)))
        for (s, d) in K.edges_iter():
            for s_i in state_dict[s]:
                for d_i in state_dict[d]:
                    if _does_respect_Xs(Xs_in_closure,
                                        self.atoms[s_i], self.atoms[d_i]):
                        self.add_edge(s_i, d_i)

    def __str__(self):
        return '(V = {}, E = {}, A = {})'.format(self.nodes(),
                                                 self.edges_iter(),
                                                 self.atoms)


def _is_non_trivial_self_fulfilling(T, C, closure):
    i_atom = next(C.__iter__())
    if len(C) > 1 or i_atom in T.next(i_atom):
        formulas = set()
        for i in C:
            formulas.update(T.atoms[i])

        for f in closure:
            if isinstance(f, CTLS.U):
                if (f in formulas) ^ (f.subformula(1) in formulas):
                    return False

        return True

    return False


def _checkE_path_formula(kripke, p_formula):
    closure = _get_closure(p_formula)
    T = _Tableu(kripke, closure=closure)

    in_ntsf = []

    for C in compute_SCCs(T):
        if _is_non_trivial_self_fulfilling(T, C, closure):
            in_ntsf.extend(C)

    T_reversed = T.get_reversed_graph()
    R = T_reversed.get_reachable_set_from(in_ntsf)

    return set([T.atoms[i].state for i in R if p_formula in T.atoms[i]])


def modelcheck(kripke, formula, parser=None, F=None):
    ''' Model checks any LTL formula on a Kripke structure.

    This method performs LTL model checking of a formula on a given
    Kripke structure.

    :param kripke: a Kripke structure.
    :type kripke: Kripke
    :param formula: the formula to model check.
    :type formula: a type castable in a LTL.Formula or a string representing
                   a LTL formula
    :param parser: a parser to parse a string into a LTL.Formula.
    :type parser: LTL.Parser
    :param F: a list of fair states
    :type F: Container
    :returns: a list of the Kripke structure states that satisfy the formula.
    '''

    if isinstance(formula, str):
        if parser is None:
            parser = Parser()
        formula = parser(formula)

    if not (isinstance(formula, CTLS.A)):
        raise TypeError('expected a LTL state formula, got {}'.format(formula))

    if not isinstance(kripke, Kripke):
        raise TypeError('expected a Kripke structure, got {}'.format(kripke))

    try:
        p_formula = LNot(formula.subformula(0))
        p_formula = p_formula.get_equivalent_restricted_formula()

        if F is not None:
            kripke = kripke.clone()

            fair_label = kripke.label_fair_states(F)

            p_formula = p_formula.get_equivalent_non_fair_formula(fair_label)
            p_formula = And(fair_label, p_formula)

        return set(kripke.states()) - _checkE_path_formula(kripke, p_formula)
    except TypeError:
        raise TypeError('expected a LTL formula, got {}'.format(formula))



import math

import numpy as np
import os
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from matplotlib.ticker import MultipleLocator


def draw_box(box_list, recorder=None, parallel=False):
    l = math.inf
    r = -math.inf
    b = math.inf
    u = -math.inf
    for rec in box_list:
        l = min(l, rec[0])
        b = min(b, rec[1])
        r = max(r, rec[2])
        u = max(u, rec[3])
    fig, ax = plt.subplots()
    width = r - l
    height = u - b
    # 画布范围
    # ax.plot([-0.2, 0.3], [-0.05, 0.1], color='white', alpha=0)
    ax.plot([l - width / 10, r + width / 10], [b - height / 10, u + height / 10], color='white', alpha=0)
    # 绘制box list
    for rec in box_list:
        ax.add_patch(
            patches.Rectangle(
                (rec[0], rec[1]),  # (x,y)
                rec[2] - rec[0],  # width
                rec[3] - rec[1],  # height
                fill=None,
                edgecolor='red'
            )
        )
    if recorder is not None:
        if parallel:
            plt.savefig(recorder.get_data_path() + '/env_parallel_BBReach.png', dpi=200)
            recorder.add_file("file/img", {
                "file_path": recorder.get_data_path() + '/env_parallel_BBReach.png',
                "info": "可达集计算图"
            })
        else:
            plt.savefig(recorder.get_data_path() + '/env_BBReach.png', dpi=200)
            recorder.add_file("file/img", {
                "file_path": recorder.get_data_path() + '/env_BBReach.png',
                "info": "可达集计算图"
            })
    plt.show(aspect='auto')


# 数组长度补齐
def box_pad(arr, dim, max_len):
    max = 10000
    min = -10000
    res = []
    pad_list = []
    half_dim = int(dim / 2)
    l = []
    for i in range(dim):
        if i < half_dim:
            l.append(max)
        else:
            l.append(min)
    pad_list.append(l)
    for item in arr:
        item_len = item.shape[0]
        while item_len < max_len:
            item = np.append(item, pad_list, axis=0)
            item_len += 1
        res.append(item)
    return np.array(res)



"""
.. module:: CTL.language
   :synopsis: Represents the CTL language.

.. moduleauthor:: Alberto Casagrande <acasagrande@units.it>
"""

import sys

from ..language import LNot
from ..language import get_alphabet, get_symbols

import trainify.validator.pyModelChecking.CTLS

CTLS = sys.modules['trainify.validator.pyModelChecking.CTLS']


class Formula(CTLS.Formula):
    '''
    A class representing CTL formulas.

    '''

    __desc__ = 'CTL formula'


class StateFormula(Formula, CTLS.StateFormula):
    '''
    A class representing CTL* state formulas.

    '''

    __desc__ = 'CTL state formula'

    def __init__(self, *phi):
        self.wrap_subformulas(phi, StateFormula)


class PathFormula(Formula, CTLS.PathFormula):
    '''
    A class representing CTL* path formulas.

    '''

    __desc__ = 'CTL path formula'

    def __init__(self, *phi):
        self.wrap_subformulas(phi, StateFormula)


class X(CTLS.X, PathFormula):
    '''
    A class representing CTL X-formulas.

    '''

    def __str__(self):
        return 'X {}'.format(self._subformula[0])


class F(CTLS.F, PathFormula):
    '''
    A class representing CTL F-formulas.

    '''

    def __str__(self):
        return 'F {}'.format(self._subformula[0])


class G(CTLS.G, PathFormula):
    '''
    A class representing CTL G-formulas.

    '''

    def __str__(self):
        return 'G {}'.format(self._subformula[0])


class U(CTLS.U, PathFormula):
    '''
    A class representing CTL U-formulas.

    '''

    pass


class R(CTLS.R, PathFormula):
    '''
    A class representing CTL R-formulas.

    '''

    pass


class AtomicProposition(CTLS.AtomicProposition, StateFormula):
    '''
    A class representing CTL atomic propositions.

    '''

    pass


class Bool(CTLS.Bool, StateFormula):
    '''
    A class representing CTL Boolean atomic propositions.

    '''

    pass


class Not(CTLS.Not, StateFormula):
    '''
    A class representing CTL negations.

    '''

    pass


class Or(CTLS.Or, StateFormula):
    '''
    A class representing CTL disjunctions.

    '''

    pass


class And(CTLS.And, StateFormula):
    '''
    A class representing CTL conjunctions.

    '''

    pass


class Imply(CTLS.Imply, StateFormula):
    '''
    A class representing CTL implications.

    '''

    pass


class A(CTLS.A, StateFormula):
    '''
    A class representing CTL A-formulas.

    '''

    def get_equivalent_restricted_formula(self):
        ''' Return a equivalent formula in the restricted syntax.

        This method returns a formula that avoids :math:`\land`,
        :math:`\\rightarrow`, :math:`A`, :math:`F`, and :math:`R` and is
        equivalent to this formula.

        :param self: this formula
        :type self: CTL.E
        :returns: a formula that avoids :math:`\land`, :math:`\\rightarrow`,
                  :math:`A`, :math:`F`, and :math:`R` and is equivalent to
                  this formula
        :rtype: CTL.StateFormula
        '''

        p_formula = self.subformula(0)
        sf0 = p_formula.subformula(0).get_equivalent_restricted_formula()
        neg_sf0 = LNot(sf0)
        if (isinstance(p_formula, CTLS.X)):
            return Not(EX(neg_sf0))

        if (isinstance(p_formula, CTLS.F)):
            return Not(EG(neg_sf0))

        if (isinstance(p_formula, CTLS.G)):
            return Not(EU(True, neg_sf0))

        sf1 = p_formula.subformula(1).get_equivalent_restricted_formula()
        neg_sf1 = LNot(sf1)
        if (isinstance(p_formula, CTLS.U)):
            return Not(Or(EU(neg_sf1, Not(Or(sf0, sf1))), EG(neg_sf1)))

        if (isinstance(p_formula, CTLS.R)):
            return Not(EU(neg_sf0, neg_sf1))

        raise TypeError('{} is not a CTL formula'.format(self))

    def get_equivalent_non_fair_formula(self, fairAP):
        p_formula = self.subformula(0)
        sf0 = p_formula.subformula(0).get_equivalent_non_fair_formula(fairAP)
        neg_sf0 = LNot(sf0)
        if (isinstance(p_formula, CTLS.X)):
            return Not(EX(And(neg_sf0, fairAP)))

        if (isinstance(p_formula, CTLS.F)):
            return Not(EG(And(neg_sf0, fairAP)))

        if (isinstance(p_formula, CTLS.G)):
            return Not(EU(True, And(neg_sf0, fairAP)))

        sf1 = p_formula.subformula(1).get_equivalent_non_fair_formula(fairAP)
        neg_sf1 = LNot(sf1)
        if (isinstance(p_formula, CTLS.U)):
            return Not(Or(EU(neg_sf1, And(Not(Or(sf0, sf1)), fairAP)),
                          EG(And(neg_sf1, fairAP))))

        if (isinstance(p_formula, CTLS.R)):
            return Not(EU(neg_sf0, And(neg_sf1, fairAP)))

        raise TypeError('{} is not a CTL formula'.format(self))

    def __str__(self):
        return 'A{}'.format(self._subformula[0])

    def __init__(self, phi):
        self.wrap_subformulas([phi], sys.modules[self.__module__].PathFormula)


class E(CTLS.E, StateFormula):
    '''
    A class representing CTL E-formulas.

    '''

    def get_equivalent_restricted_formula(self):
        ''' Return a equivalent formula in the restricted syntax.

        This method returns a formula that avoids :math:`\land`,
        :math:`\\rightarrow`, :math:`A`, :math:`F`, and :math:`R` and is
        equivalent to this formula.

        :param self: this formula
        :type self: CTL.E
        :returns: a formula that avoids :math:`\land`, :math:`\\rightarrow`,
                  :math:`A`, :math:`F`, and :math:`R` and is equivalent to
                  this formula
        :rtype: CTL.StateFormula
        '''

        p_formula = self.subformula(0)
        sf0 = p_formula.subformula(0).get_equivalent_restricted_formula()
        if (isinstance(p_formula, CTLS.X)):
            return EX(sf0)

        if (isinstance(p_formula, CTLS.F)):
            return EU(True, sf0)

        if (isinstance(p_formula, CTLS.G)):
            return EG(sf0)

        sf1 = p_formula.subformula(1).get_equivalent_restricted_formula()
        if (isinstance(p_formula, CTLS.U)):
            return EU(sf0, sf1)

        if (isinstance(p_formula, CTLS.R)):
            neg_sf1 = LNot(sf1)
            neg_sf0 = LNot(sf0)

            return Or(EU(sf1, Not(Or(neg_sf0, neg_sf1))), EG(sf1))

        raise TypeError('{} is not a CTL formula'.format(self))

    def get_equivalent_non_fair_formula(self, fairAP):
        p_formula = self.subformula(0)
        sf0 = p_formula.subformula(0).get_equivalent_non_fair_formula(fairAP)
        if (isinstance(p_formula, CTLS.X)):
            return EX(And(sf0, fairAP))

        if (isinstance(p_formula, CTLS.F)):
            return EU(True, And(sf0, fairAP))

        if (isinstance(p_formula, CTLS.G)):
            return EG(And(sf0, fairAP))

        sf1 = p_formula.subformula(1).get_equivalent_non_fair_formula(fairAP)
        if (isinstance(p_formula, CTLS.U)):
            return EU(sf0, And(sf1, fairAP))

        if (isinstance(p_formula, CTLS.R)):
            neg_sf1 = LNot(sf1)
            neg_sf0 = LNot(sf0)

            return Or(EU(sf1, And(Not(Or(neg_sf0, neg_sf1))), fairAP),
                      EG(And(sf1, fairAP)))

        raise TypeError('{} is not a CTL formula'.format(self))

    def __str__(self):
        return 'E{}'.format(self._subformula[0])

    def __init__(self, phi):
        self.wrap_subformulas([phi], sys.modules[self.__module__].PathFormula)


def AX(psi):
    ''' A shortcut to build :math:`A(X(\psi))`.

    This method returns the formula :math:`A(X(\psi))` where :math:`\psi` is
    the method parameter.

    :param psi: a state formula
    :type psi: CTL.StateFormula
    :returns: the formula :math:`A(X(\psi))`
    :rtype: CTL.StateFormula
    '''
    return A(X(psi))


def EX(psi):
    ''' A shortcut to build :math:`E(X(\psi))`.

    This method returns the formula :math:`E(X(\psi))` where :math:`\psi` is
    the method parameter.

    :param psi: a state formula
    :type psi: CTL.StateFormula
    :returns: the formula :math:`E(X(\psi))`
    :rtype: CTL.StateFormula
    '''
    return E(X(psi))


def AF(psi):
    ''' A shortcut to build :math:`E(X(\psi))`.

    This method returns the formula :math:`E(X(\psi))` where :math:`\psi` is
    the method parameter.

    :param psi: a state formula
    :type psi: CTL.StateFormula
    :returns: the formula :math:`E(X(\psi))`
    :rtype: CTL.StateFormula
    '''
    return A(F(psi))


def EF(psi):
    ''' A shortcut to build :math:`E(F(\psi))`.

    This method returns the formula :math:`E(F(\psi))` where :math:`\psi` is
    the method parameter.

    :param psi: a state formula
    :type psi: CTL.StateFormula
    :returns: the formula :math:`E(F(\psi))`
    :rtype: CTL.StateFormula
    '''
    return E(F(psi))


def AG(psi):
    ''' A shortcut to build :math:`A(G(\psi))`.

    This method returns the formula :math:`A(G(\psi))` where :math:`\psi` is
    the method parameter.

    :param psi: a state formula
    :type psi: CTL.StateFormula
    :returns: the formula :math:`A(G(\psi))`
    :rtype: CTL.StateFormula
    '''
    return A(G(psi))


def EG(psi):
    ''' A shortcut to build :math:`E(G(\psi))`.

    This method returns the formula :math:`E(G(\psi))` where :math:`\psi` is
    the method parameter.

    :param psi: a state formula
    :type psi: CTL.StateFormula
    :returns: the formula :math:`E(G(\psi))`
    :rtype: CTL.StateFormula
    '''
    return E(G(psi))


def AU(psi, phi):
    ''' A shortcut to build :math:`A(U(\psi, \phi))`.

    This method returns the formula :math:`A(U(\psi, \phi))` where
    :math:`\psi` and :math:`\phi` are the method parameters.

    :param psi: a state formula
    :type psi: CTL.StateFormula
    :param phi: a state formula
    :type phi: CTL.StateFormula
    :returns: the formula :math:`A(U(\psi, \phi))`
    :rtype: CTL.StateFormula
    '''
    return A(U(psi, phi))


def EU(psi, phi):
    ''' A shortcut to build :math:`E(U(\psi, \phi))`.

    This method returns the formula :math:`E(U(\psi, \phi))` where
    :math:`\psi` and :math:`\phi` are the method parameters.

    :param psi: a state formula
    :type psi: CTL.StateFormula
    :param phi: a state formula
    :type phi: CTL.StateFormula
    :returns: the formula :math:`E(U(\psi, \phi))`
    :rtype: CTL.StateFormula
    '''
    return E(U(psi, phi))


def AR(psi, phi):
    ''' A shortcut to build :math:`A(R(\psi, \phi))`.

    This method returns the formula :math:`A(R(\psi, \phi))` where
    :math:`\psi` and :math:`\phi` are the method parameters.

    :param psi: a state formula
    :type psi: CTL.StateFormula
    :param phi: a state formula
    :type phi: CTL.StateFormula
    :returns: the formula :math:`A(R(\psi, \phi))`
    :rtype: CTL.StateFormula
    '''
    return A(R(psi, phi))


def ER(psi, phi):
    ''' A shortcut to build :math:`E(R(\psi, \phi))`.

    This method returns the formula :math:`E(R(\psi, \phi))` where
    :math:`\psi` and :math:`\phi` are the method parameters.

    :param psi: a state formula
    :type psi: CTL.StateFormula
    :param phi: a state formula
    :type phi: CTL.StateFormula
    :returns: the formula :math:`E(R(\psi, \phi))`
    :rtype: CTL.StateFormula
    '''
    return E(R(psi, phi))


alphabet = get_alphabet(__name__)
symbols = get_symbols(alphabet)



import ast

from .BDD import BDDNode
from .BDD import apply as BDDapply
from .ordering import *


def parse_args(args_node):
    return Ordering([id(arg) for arg in args_node.args])


def parse_name(ordering, node):
    if node.id == 'True':
        return OBDD(BDDNode(True), ordering)
    if node.id == 'False':
        return OBDD(BDDNode(False), ordering)

    return OBDD(BDDNode(node.id, BDDNode(False), BDDNode(True)), ordering)


def parse_binary_unary_op(ordering, node):
    if (isinstance(node.op, ast.Not) or isinstance(node.op, ast.Invert)):

        return ~parse_binary_expr(ordering, node.operand)


def parse_binary_op(ordering, node):
    if isinstance(node.op, ast.BitAnd):
        return (parse_binary_expr(ordering, node.left) &
                parse_binary_expr(ordering, node.right))

    if isinstance(node.op, ast.BitOr):
        return (parse_binary_expr(ordering, node.left) |
                parse_binary_expr(ordering, node.right))

    raise SyntaxError('expected a binary operator, got a ' +
                      '{}'.format(node.op.__class__))


def parse_binary_binary_op(ordering, node):
    if isinstance(node.op, ast.And):
        result = OBDD(BDDNode(True), ordering)
        for arg in node.values:
            result = result & parse_binary_expr(ordering, arg)

        return result

    if isinstance(node.op, ast.Or):
        result = OBDD(BDDNode(False), ordering)
        for arg in node.values:
            result = result | parse_binary_expr(ordering, arg)

        return result

    raise SyntaxError('expected a binary operator, got a ' +
                      '{}'.format(node.op.__class__))


def parse_binary_expr(ordering, node):
    if isinstance(node, ast.BinOp):
        return parse_binary_op(ordering, node)

    if isinstance(node, ast.BoolOp):
        return parse_binary_binary_op(ordering, node)

    if isinstance(node, ast.UnaryOp):
        return parse_binary_unary_op(ordering, node)

    if isinstance(node, ast.Name):
        return parse_name(ordering, node)

    if isinstance(node, ast.Num):
        if node.n in [0, 1]:
            return OBDD(BDDNode(node.n), ordering)

        raise SyntaxError('expected a binary expression, got number ' +
                          '{}'.format(node.n))

    raise SyntaxError('expected a binary expression, got a ' +
                      '{}'.format(node.__class__))


class BinaryParser(object):
    '''
    A class to represent parsers of binary functions.

    The objects of this class can parse strings representing binary functions
    are return the corresponding OBDD.

    '''

    def __init__(self, ordering):
        ''' Initialize a binary parser.

        :param ordering: an ordering for the variables to be parsed
        :type ordering: Ordering
        '''
        if not isinstance(ordering, Ordering):
            try:
                ordering = Ordering(ordering)
            except TypeError:
                raise TypeError('expected an Ordering of variables, got ' +
                                '{}'.format(ordering))

        self.ordering = ordering

    @staticmethod
    def parse_function(binary_funct):
        ''' Parse a binary function.

        :param ordering: an ordering for the variables to be parsed
        :type ordering: Ordering
        '''
        st = ast.parse(binary_funct)

        try:
            args_node = st.body[0].value.args

            body_node = st.body[0].value.body

        except Exception:
            raise SyntaxError('expected a binary lambda function, got ' +
                              '{}'.format(binary_funct))

        ordering = parse_args(args_node)
        return parse_binary_expr(ordering, body_node)

    def parse(self, binary_expr):
        st = ast.parse(binary_expr)
        try:
            binary_expr_node = st.body[0].value
        except Exception:
            raise SyntaxError('expected a binary expression, got ' +
                              '{}'.format(binary_expr))

        return parse_binary_expr(self.ordering, binary_expr_node)


class OBDD(object):
    '''
    A class to represent Ordered Binary Decision Diagrams (OBDDs).
    '''

    def __init__(self, bfunct, ordering=None, check_ordering=True):
        ''' Initialize an OBDD.

        :param bfunct: either a BDDNode,  a string that represents a binary
                expression,  or a string that represent a binary function in
                lambda notation
        :type bfunct: str or function
        :param ordering: an ordering for the variables to be parsed
        :type ordering: Ordering
        :param check_ordering: a binary flag: if bfunct is a BDDNode and this
                               flag is set to True,  then this method check
                               whether the BDDNode respects the ordering
        :type check_ordering: bool
        '''
        if ordering is None:
            obdd = BinaryParser.parse_function(bfunct)
            self.ordering = obdd.ordering
            self.root = obdd.root

            return

        if not isinstance(ordering, Ordering):
            try:
                ordering = Ordering(ordering)
            except TypeError:
                raise TypeError('expected an Ordering of variables, got ' +
                                '{}'.format(ordering))

        self.ordering = ordering

        if isinstance(bfunct, BDDNode):
            if check_ordering and not bfunct.respect_ordering(ordering):
                raise ValueError('the BDDNode {} '.format(bfunct) +
                                 'does not respect {}'.format(ordering))

            self.root = bfunct
        else:
            if isinstance(bfunct, str):
                bparser = BinaryParser(ordering)
                self.root = (bparser.parse(bfunct)).root
            else:
                raise TypeError('expected a BDDNode or a str, got ' +
                                '{}'.format(bfunct))

    def restrict(self, var, value):
        ''' Partially evaluate the binary function encoded by an OBDD.

        :param var: name of the variable to be set
        :type var: str
        :param value: value to be set
        :type value: bool or int
        :returns: the OBDD representing the partial evaluation of :math:`f`
                  with :param var: set to :param value: where :math:`f` is
                  the function encoded by current object
        :rtype: OBDD
        '''
        return OBDD(self.root.restrict(var, value), self.ordering)

    def variables(self):
        ''' Return the variables in an OBDD.

        :returns: the set of the variables represented in the OBDD
        :rtype: set
        '''
        return self.root.variables()

    def __eq__(self, A):
        ''' Check whether two OBDD are the same.

        :param A: either a BDDNode,  an OBDD,  a binary,  or a Boolean value
        :type A: BDDNode,  OBDD,  binary,  or Boolean value
        :returns: True if the two OBDD are the same and respect the same
                  ordering,  False otherwise
        :rtype: bool
        '''
        if isinstance(A, BDDNode):
            return self == OBDD(A, self.ordering)

        if isinstance(A, OBDD):
            return self.root is A.root and self.ordering == A.ordering

        if A in set([0, 1, True, False]):
            return self == OBDD(BDDNode(A), self.ordering)

        raise TypeError('expected an OBDD, got %s' % (A))

    def __req__(self, A):
        ''' Check whether two OBDD are the same.

        :param A: either a BDDNode,  an OBDD,  a binary,  or a Boolean value
        :type A: BDDNode,  OBDD,  binary,  or Boolean value
        :returns: True if the two OBDD are the same and respect the same
                  ordering,  False otherwise
        :rtype: bool
        '''
        return self == A

    def apply(self, operator, B):
        ''' Apply a binary binary operator to two OBDD.

        :param operator: a binary boolean operator
        :type operator: function
        :param A: an OBDD
        :type A: OBDD
        :param B: an OBDD
        :type B: OBDD
        :returns: The OBDD obtained by applying the operator to the two OBDD
        :rtype: OBDD
        '''

        if not isinstance(B, OBDD):
            raise TypeError('expected an OBDD, got {}'.format(B))

        if self.ordering != B.ordering:
            raise RuntimeError('Unsupported operation: {}'.format(self) +
                               ' and {} '.format(B) +
                               'have different variable ordering')

        result_cache = dict()

        bdd = BDDapply(operator, self.root, B.root, self.ordering,
                       result_cache)
        return OBDD(bdd, self.ordering, check_ordering=False)

    def __and__(self, A):
        ''' Build the conjunction of two OBDD.

        :param A: an OBDD
        :type A: OBDD
        :returns: the OBDD that represents the logical conjunction of
                  the two OBDD
        :rtype: OBDD
        '''
        return self.apply((lambda a, b: a and b), A)

    def __or__(self, A):
        ''' Build the non-exclusive disjunction of two OBDD.

        :param A: an OBDD
        :type A: OBDD
        :returns: the OBDD that represents the logical non-exclusive
                  disjunction of the two OBDD
        :rtype: OBDD
        '''
        return self.apply((lambda a, b: a or b), A)

    def __xor__(self, A):
        ''' Build the exclusive disjunction of two OBDD.

        :param A: an OBDD
        :type A: OBDD
        :returns: the OBDD that represents the logical exclusive disjunction
                  of the two OBDD
        :rtype: OBDD
        '''
        return self.apply((lambda a, b: a ^ b), A)

    def __invert__(self):
        ''' Build the negation of an OBDD.

        :param A: an OBDD
        :type A: OBDD
        :returns: the OBDD that represents the logical negation of the OBDD
        :rtype: OBDD
        '''
        return OBDD(~self.root, self.ordering)

    def __str__(self):
        ''' Return a string that represents an OBDD

        :returns: a string that represents the OBDD
        :rtype: str
        '''
        if isinstance(self.ordering, ListOrdering):
            result = 'lambda'
            sep = ' '
            for var in self.ordering.get_list():
                result += '{}{}'.format(sep, var)
                sep = ','

            return result+': {}'.format(self.root)

        return '{}'.format(self.root)

    def __repr__(self):
        ''' Return a string that represents an OBDD

        :returns: a string that represents the OBDD
        :rtype: str
        '''
        return '{}'.format(str(self))



import torch
from scipy.optimize import minimize, Bounds
import math
import numpy as np

from trainify.utils import str_to_list


class PendulumEnv():

    def __init__(self, divide_tool, network):
        self.initial_state = [1, 0, 0.1]
        self.initial_state_region = None

        # proposition_list, limited_count, limited_depth, atomic_propositions, formula,
        #                  get_abstract_state, get_abstract_state_label, get_abstract_state_hash, rtree
        self.proposition_list = []
        self.limited_count = 500000
        self.limited_depth = 800
        self.atomic_propositions = ['safe']
        self.formula = 'not(A(G(safe)))'

        self.max_speed = 8
        self.max_torque = 2.
        self.dt = .05
        self.g = 10.0
        self.m = 1.
        self.l = 1.

        self.action = 0.0
        self.epsilon = 0.0

        self.divide_tool = divide_tool
        self.network = network

    # state, inteval, range
    def is_done(self, s):

        done = bool(
            s[0] <= 0.0
        )
        return done

    def get_abstract_state(self, s):
        return self.divide_tool.get_abstract_state(s)

    def get_abstract_state_label(self, abstract_state, cnt):
        state_list = str_to_list(abstract_state)
        if state_list[0] <= 0:
            # if cnt <= 5:
            #     print('bad state-----------', state_list)
            return []
        return ['safe']

    def get_abstract_state_hash(self, abstract_state):
        return str(abstract_state)

    # Define related functions for optimization in SciPy
    def thdot_minimum(self, x):
        # x[1] + (-3 * self.g / (2 * self.l) * np.sin(th + np.pi) + 3. / (m * l ** 2) * u) *
        return x[1] + (3 * self.g / (2 * self.l) * math.sin(x[0]) + 3. / (self.m * self.l ** 2) * self.action) * self.dt

    def thdot_maximum(self, x):
        return - self.thdot_minimum(x)

    def cos_minimum(self, x):
        t = self.thdot_minimum(x)
        # cos(newth) = cos(th + t * dt) = cos(th) * cos(t * dt)  - sin(th) * sin(t * dt)
        cosin = math.cos(x[0] + t * self.dt)
        return cosin

    def cos_maximum(self, x):
        return - self.cos_minimum(x)

    def sin_minimum(self, x):
        t = self.thdot_minimum(x)
        # sin(newth) = sin(th + t * dt) = sin(th) * cos(t * dt)  + cos(th) * cos(t * dt)
        sine = math.sin(x[0] + t * self.dt)
        return sine

    def sin_maximum(self, x):
        return - self.sin_minimum(x)

    # Method must be implemented by users
    def get_next_states(self, current):
        cs = current
        try:
            current = str_to_list(current)
        except:
            print(current)
            exit(0)
        if current[0] <= 0:
            return [cs]

        s0 = torch.tensor(current, dtype=torch.float).unsqueeze(0)
        action = self.network(s0).squeeze(0).detach().numpy()

        self.action = np.clip(action[0], -self.max_torque, self.max_torque)

        tmp1 = np.clip(current[1], -1, 1)
        tmp2 = np.clip(current[4], -1, 1)
        min_th = math.asin(tmp1)
        max_th = math.asin(tmp2)
        x0 = [min_th, current[2]]

        bounds = Bounds([min_th, current[2]], [max_th, current[5]])
        thdot_left = minimize(self.thdot_minimum, x0, method='SLSQP', bounds=bounds)
        thdot_left = self.thdot_minimum(thdot_left.x)
        thdot_right = minimize(self.thdot_maximum, x0, method='SLSQP', bounds=bounds)
        thdot_right = - self.thdot_maximum(thdot_right.x)

        cos_left = minimize(self.cos_minimum, x0, method='SLSQP', bounds=bounds)
        cos_left = self.cos_minimum(cos_left.x)
        cos_right = minimize(self.cos_maximum, x0, method='SLSQP', bounds=bounds)
        cos_right = - self.cos_maximum(cos_right.x)

        if cos_left < -1.0 or cos_right > 1.0 or cos_left > cos_right:
            print('pendulum/scenario_transition/cosin: ', current)
            print(cos_left, cos_right)
            exit(0)

        sin_left = minimize(self.sin_minimum, x0, method='SLSQP', bounds=bounds)
        sin_left = self.sin_minimum(sin_left.x)
        sin_right = minimize(self.sin_maximum, x0, method='SLSQP', bounds=bounds)
        sin_right = - self.sin_maximum(sin_right.x)

        if sin_left < -1.0 or sin_right > 1.0 or sin_left > sin_right:
            print('pendulum/scenario_transition/sine: ', current)
            print(sin_left, sin_right)
            exit(0)

        thdot_left = np.clip(thdot_left, -self.max_speed, self.max_speed)
        thdot_right = np.clip(thdot_right, -self.max_speed, self.max_speed)

        next_bounds = [cos_left, sin_left, thdot_left, cos_right, sin_right, thdot_right]
        next_states = self.divide_tool.intersection(next_bounds)
        return next_states

    def get_low_dim_state(self, state):
        s = str_to_list(state)
        res = [s[0], s[1], s[3], s[4]]
        obj_str = ','.join([str(_) for _ in res])
        return obj_str
